"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from typing import Optional
from urllib.parse import urlparse

import boto3
import pandas as pd
from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator, DagRun, TaskInstance
from google.cloud.storage import Client
from smart_open import open

from astro.sql.operators.temp_hooks import TempPostgresHook, TempSnowflakeHook
from astro.sql.table import Table
from astro.utils.cloud_storage_creds import gcs_client, s3fs_creds
from astro.utils.task_id_helper import get_task_id


class SaveFile(BaseOperator):
    """Write SQL table to csv/parquet on local/S3/GCS.

    :param output_file_path: Path and name of table to create.
    :type output_file_path: str
    :param table: Input table name.
    :type table: str
    :param input_conn_id: Database connection id.
    :type input_conn_id: str
    :param output_conn_id: File system connection id (if S3 or GCS).
    :type output_conn_id: str
    :param overwrite: Overwrite file if exists. Default False.
    :type overwrite: bool
    :param output_file_format: file formats, valid values csv/parquet. Default: 'csv'.
    :type output_file_format: str
    """

    def __init__(
        self,
        table="",
        output_file_path="",
        input_table: Table = None,
        output_conn_id=None,
        output_file_format="csv",
        overwrite=None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.table = table
        self.output_file_path = output_file_path
        self.input_table = input_table
        self.output_conn_id = output_conn_id
        self.overwrite = overwrite
        self.output_file_format = output_file_format
        self.kwargs = kwargs

    def execute(self, context):
        """Write SQL table to csv/parquet on local/S3/GCS.

        Infers SQL database type based on connection.
        """

        # Infer db type from `input_conn_id`.
        input_table = self.input_table
        conn_type = BaseHook.get_connection(input_table.conn_id).conn_type

        # Select database Hook based on `conn` type
        input_hook = {
            "postgres": TempPostgresHook(
                postgres_conn_id=input_table.conn_id, schema=input_table.database
            ),
            "snowflake": TempSnowflakeHook(
                snowflake_conn_id=input_table.conn_id,
                database=input_table.database,
                schema=input_table.schema,
                warehouse=input_table.warehouse,
            ),
        }.get(conn_type, None)

        # Load table from SQL db.
        df = pd.read_sql(
            f"SELECT * FROM {input_table.table_name}",
            con=input_hook.get_sqlalchemy_engine(),
        )

        # Write file if overwrite == True or if file doesn't exist.
        if self.overwrite == True or not self.file_exists(
            self.output_file_path, self.output_conn_id
        ):
            self.agnostic_write_file(df, self.output_file_path, self.output_conn_id)
        else:
            raise FileExistsError

    def file_exists(self, output_file_path, output_conn_id=None):
        transport_params = {
            "s3": s3fs_creds,
            "gs": gcs_client,
            "": lambda: None,
        }[urlparse(output_file_path).scheme]()
        try:
            with open(output_file_path, mode="r", transport_params=transport_params):
                return True
        except IOError:
            return False

    def agnostic_write_file(self, df, output_file_path, output_conn_id=None):
        """Write dataframe to csv/parquet files formats

        Select output file format based on param output_file_format to class.
        """
        transport_params = {
            "s3": s3fs_creds,
            "gs": gcs_client,
            "": lambda: None,
        }[urlparse(output_file_path).scheme]()

        serialiser = {
            "parquet": df.to_parquet,
            "csv": df.to_csv,
            "json": df.to_json,
            "ndjson": df.to_json,
        }
        serialiser_params = {
            "csv": {"index": False},
            "json": {"orient": "records"},
            "ndjson": {"orient": "records", "lines": True},
        }
        with open(
            output_file_path, mode="wb", transport_params=transport_params
        ) as stream:
            serialiser[self.output_file_format](
                stream, **serialiser_params.get(self.output_file_format, {})
            )

    @staticmethod
    def create_table_name(context):
        ti: TaskInstance = context["ti"]
        dag_run: DagRun = ti.get_dagrun()
        return f"{dag_run.dag_id}_{ti.task_id}_{dag_run.id}"


def save_file(
    output_file_path,
    table=None,
    input_table=None,
    output_conn_id=None,
    overwrite=False,
    output_file_format="csv",
    task_id=None,
    **kwargs,
):
    """Convert SaveFile into a function. Returns XComArg.

    Returns an XComArg object.

    :param output_file_path: Path and name of table to create.
    :type output_file_path: str
    :param table: Input table name.
    :type table: str
    :param input_conn_id: Database connection id.
    :type input_conn_id: str
    :param output_conn_id: File system connection id (if S3 or GCS).
    :type output_conn_id: str
    :param overwrite: Overwrite file if exists. Default False.
    :type overwrite: bool
    :param task_id: task id, optional.
    :type task_id: str
    """

    task_id = (
        task_id if task_id is not None else get_task_id("save_file", output_file_path)
    )

    return SaveFile(
        task_id=task_id,
        output_file_path=output_file_path,
        table=table,
        input_table=input_table,
        output_conn_id=output_conn_id,
        overwrite=overwrite,
        output_file_format=output_file_format,
    ).output



import os
import re
from collections import defaultdict
from typing import Dict

import frontmatter
from airflow.decorators.base import get_unique_task_id
from airflow.decorators.task_group import task_group
from airflow.exceptions import AirflowException
from airflow.models.xcom_arg import XComArg

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table, TempTable


@task_group()
def render(path, **kwargs):
    # raise AirflowException(f"Failed because cwd is {os.listdir(path)}, {os.}")
    files = [
        f
        for f in os.listdir(path)
        if os.path.isfile(os.path.join(path, f)) and f.endswith(".sql")
    ]
    template_dict = kwargs
    op_kwargs = {}
    # Parse all of the SQL files in this directory
    for filename in files:
        with open(os.path.join(path, filename), "r") as f:
            front_matter_opts = frontmatter.loads(f.read()).to_dict()
            sql = front_matter_opts.pop("content")
            templated_names = find_templated_fields(sql)
            parameters = {y: None for y in templated_names}
            if front_matter_opts.get("template_vars"):
                template_variables = front_matter_opts.pop("template_vars")
                sql = wrap_template_variables(sql, template_variables)
                parameters.update({v: None for k, v in template_variables.items()})
            if front_matter_opts.get("output_table"):
                out_table_dict = front_matter_opts.pop("output_table")
                if out_table_dict.get("table_name"):
                    op_kwargs = {"output_table": Table(**out_table_dict)}
                else:
                    op_kwargs = {"output_table": TempTable(**out_table_dict)}

            p = ParsedSqlOperator(
                sql=sql,
                parameters=parameters,
                file_name=filename,
                op_kwargs=op_kwargs,
                **front_matter_opts,
            )
            template_dict[filename.replace(".sql", "")] = p.output

    # Add the XComArg to the parameters to create dependency
    for filename in files:
        current_operator = template_dict[filename.replace(".sql", "")].operator
        for param in current_operator.parameters:
            if not template_dict.get(param):
                raise AirflowException(
                    f"Table {param} does not exist for file {filename}"
                )
            current_operator.parameters[param] = template_dict[param]
            # due to an edge case in XComArg, we need to explicitly set dependencies here
            if type(template_dict[param]) == XComArg:
                template_dict[param].operator >> current_operator
    ret = []
    for f in template_dict.values():
        ret.append(f)
    return template_dict


def find_templated_fields(file_string):
    return [y[1:-1] for y in re.findall(r"\{[^}]*\}", file_string) if "{{" not in y]


def wrap_template_variables(sql, template_vars):
    words = sql.split(" ")
    fixed_words = [
        "{" + template_vars.get(w) + "}" if template_vars.get(w) else w for w in words
    ]
    return " ".join(fixed_words)


class ParsedSqlOperator(SqlDecoratoratedOperator):
    template_fields = ("sql", "parameters")

    def _table_exists_in_db(self, conn: str, table_name: str):
        pass

    def handle_dataframe_func(self, input_table):
        pass

    def __init__(
        self,
        sql,
        parameters,
        file_name,
        op_kwargs={},
        **kwargs,
    ):
        self.sql = sql
        self.parameters = parameters
        task_id = get_unique_task_id(file_name.replace(".sql", ""))

        def null_function():
            return sql, parameters

        super().__init__(
            raw_sql=False,
            task_id=task_id,
            sql=sql,
            op_args=(),
            parameters=parameters,
            python_callable=null_function,
            op_kwargs=op_kwargs,
            **kwargs,
        )

    def execute(self, context: Dict):
        super().execute(context)



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from urllib.parse import quote_plus

from airflow.providers.postgres.hooks.postgres import PostgresHook
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook


class TempSnowflakeHook(SnowflakeHook):
    """
    Temporary class to get around a bug in the snowflakehook when creating URIs
    """

    def get_uri(self) -> str:
        """Override DbApiHook get_uri method for get_sqlalchemy_engine()"""
        conn_config = self._get_conn_params()
        uri = (
            "snowflake://{user}:{password}@{account}/{database}/{schema}"
            "?warehouse={warehouse}&role={role}&authenticator={authenticator}"
        )
        return uri.format(**conn_config)


class TempPostgresHook(PostgresHook):
    """
    Temporary class to get around a bug in the snowflakehook when creating URIs
    """

    def get_uri(self) -> str:
        """
        Extract the URI from the connection.

        :return: the extracted uri.
        """
        conn = self.get_connection(getattr(self, self.conn_name_attr))
        login = ""
        if conn.login:
            login = f"{quote_plus(conn.login)}:{quote_plus(conn.password)}@"
        host = conn.host
        if conn.port is not None:
            host += f":{conn.port}"
        uri = f"postgresql://{login}{host}/"
        if self.schema:
            uri += self.schema
        return uri



from distutils import log as logger
from os import name, stat
from typing import Dict, List

from airflow.hooks.base import BaseHook
from airflow.providers.postgres.hooks.postgres import PostgresHook
from psycopg2 import sql
from sqlalchemy.sql.expression import table
from sqlalchemy.sql.functions import Function
from sqlalchemy.sql.schema import Table

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table
from astro.utils.snowflake_merge_func import is_valid_snow_identifier


class Check:
    def __init__(
        self,
        name: str,
        expression: str,
        threshold: float = 0,
    ) -> None:
        self.name = name
        self.expression = expression
        self.threshold = threshold

    def get_expression(self, conn_type: str):
        return {
            "postgres": self.get_postgres_expression(),
            "snowflake": self.get_snowflake_expression(),
        }[conn_type]

    def get_postgres_expression(self):
        return sql.SQL("CASE WHEN {expression} THEN 0 ELSE 1 END AS {name}").format(
            expression=sql.SQL(self.expression), name=sql.Identifier(self.name)
        )

    def get_postgres_result(self):
        return sql.SQL("CAST(SUM({name}) as float) / COUNT(*) as {result}").format(
            name=sql.Identifier(self.name), result=sql.Identifier(self.name + "_result")
        )

    def get_snowflake_expression(self):
        if not is_valid_snow_identifier(name):
            raise ValueError("Not a valid snowflake identifier {}".format(name))
        return "CASE WHEN {}  THEN 0 ELSE 1 END AS {}".format(
            self.expression, self.name
        )

    def get_snowflake_result(self):
        if not is_valid_snow_identifier(name):
            raise ValueError("Not a valid snowflake identifier {}".format(name))
        return 'CAST(SUM({}) as float) / COUNT(*) as "{}"'.format(
            self.name, self.name + "_result"
        )


class AgnosticBooleanCheck(SqlDecoratoratedOperator):
    def __init__(
        self,
        checks: List[Check],
        table: Table,
        max_rows_returned: int,
        **kwargs,
    ):
        """
        :param table: table name
        :type table: str
        :param checks: check class object, which represent boolean expression
        :type checks: Check
        :param max_rows_returned: number of row returned if the check fails.
        :type max_rows_returned: int
        """

        self.table = table
        self.max_rows_returned = max_rows_returned
        self.conn_id = table.conn_id
        self.checks = checks
        self.database = table.database

        task_id = table.table_name + "_" + "boolean_check"

        def null_function():
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            conn_id=table.conn_id,
            database=table.database,
            schema=table.schema,
            warehouse=table.warehouse,
            task_id=task_id,
            op_args=(),
            python_callable=null_function,
            **kwargs,
        )

    def execute(self, context: Dict):
        conn_type = BaseHook.get_connection(self.conn_id).conn_type  # type: ignore
        postgres = "postgres"
        snowflake = "snowflake"
        try:
            execute_boolean_checks = {
                postgres: AgnosticBooleanCheck.execute_postgres_boolean_checks,
                snowflake: AgnosticBooleanCheck.execute_snowflake_boolean_checks,
            }[conn_type]
            get_failed_checks = {
                postgres: self.get_postgres_failed_checks,
                snowflake: self.get_snowflake_failed_checks,
            }[conn_type]
            prep_results = {
                postgres: self.postgres_prep_results,
                snowflake: self.snowflake_prep_results,
            }[conn_type]
        except KeyError:
            raise ValueError(f"Please specify a postgres or snowflake conn id.")

        self.handler = lambda curr: curr.fetchall()

        self.sql, self.parameters = execute_boolean_checks(
            self.table.table_name, self.checks, self.conn_id  # type: ignore
        )

        results = super().execute(context)
        failed_checks_names, failed_checks_index = get_failed_checks(results)
        if len(failed_checks_index) > 0:
            self.sql, self.parameters = prep_results(failed_checks_index)
            failed_rows = super().execute(context)
            logger.error("Failed rows {}".format(failed_rows))
            raise ValueError(
                "Some of the check(s) have failed {}".format(
                    ",".join(failed_checks_names)
                )
            )

        return table

    def get_postgres_failed_checks(self, results):
        failed_check_name = []
        failed_check_index = []

        for index in range(len(self.checks)):
            if self.checks[index].threshold < results[0][index]:
                failed_check_name.append(self.checks[index].name)
                failed_check_index.append(index)
        return failed_check_name, failed_check_index

    def get_snowflake_failed_checks(self, results):
        failed_check_name = []
        failed_check_index = []

        for index in range(len(self.checks)):
            if (
                self.checks[index].threshold
                < results[0][self.checks[index].name + "_result"]
            ):
                failed_check_name.append(self.checks[index].name)
                failed_check_index.append(index)
        return failed_check_name, failed_check_index

    @staticmethod
    def execute_postgres_boolean_checks(table: str, checks: List[Check], conn_id: str):
        statement = (
            "SELECT {results} FROM (SELECT {expressions} From {table}) as temp_table"
        )
        expressions = sql.SQL(",").join(
            [check.get_postgres_expression() for check in checks]
        )
        results = sql.SQL(",").join([check.get_postgres_result() for check in checks])

        hook = PostgresHook(postgres_conn_id=conn_id)

        statement = (
            sql.SQL(statement)
            .format(
                results=results, expressions=expressions, table=sql.Identifier(table)
            )
            .as_string(hook.get_conn())
        )
        return statement, {}

    @staticmethod
    def execute_snowflake_boolean_checks(table: str, checks: List[Check], conn_id: str):
        statement = "SELECT {results} FROM (SELECT {expressions} From Identifier(%(table)s) as temp_table)"

        statement = statement.replace("{table}", table)
        statement = statement.replace(
            "{expressions}",
            ",".join([check.get_snowflake_expression() for check in checks]),
        )
        statement = statement.replace(
            "{results}", ",".join([check.get_snowflake_result() for check in checks])
        )
        return statement, {"table": table}

    def postgres_prep_results(self, results):
        statement = "SELECT * from {table} WHERE {checks} LIMIT {max_rows_returned}"
        failed_checks = [sql.SQL(self.checks[index].expression) for index in results]
        hook = PostgresHook(postgres_conn_id=self.conn_id)
        statement = (
            sql.SQL(statement)
            .format(
                table=sql.Identifier(self.table.table_name),
                checks=sql.SQL("AND").join(failed_checks),
                max_rows_returned=sql.SQL(str(self.max_rows_returned)),
            )
            .as_string(hook.get_conn())
        )
        return statement, {}

    def snowflake_prep_results(self, results):
        statement = (
            "SELECT * from Identifier(%(table)s) WHERE {expressions} LIMIT {limit}"
        )
        statement = statement.replace(
            "{expressions}",
            "AND".join([self.checks[index].expression for index in results]),
        )
        statement = statement.replace("{limit}", str(self.max_rows_returned))
        return statement, {"table": self.table}


def wrap_identifier(inp):
    return "Identifier(%(" + inp + ")s)"


def boolean_check(table: Table, checks: List[Check] = [], max_rows_returned: int = 100):
    """
    :param table: table name
    :type table: str
    :param checks: check class object, which represent boolean expression
    :type checks: Check
    :param max_rows_returned: number of row returned if the check fails.
    :type max_rows_returned: int
    """

    return AgnosticBooleanCheck(
        table=table, checks=checks, max_rows_returned=max_rows_returned
    )



from distutils import log as logger
from os import name, stat
from typing import Dict, List, Set

from airflow.hooks.base import BaseHook
from airflow.providers.postgres.hooks.postgres import PostgresHook
from psycopg2 import sql
from sqlalchemy.sql.expression import table
from sqlalchemy.sql.functions import Function
from sqlalchemy.sql.schema import Table

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table
from astro.utils.snowflake_merge_func import (
    is_valid_snow_identifier,
    is_valid_snow_identifiers,
)


class OutlierCheck:
    def __init__(
        self,
        name: str,
        columns_map: map,
        accepted_std_div: int,
        threshold: float,
    ) -> None:
        self.name = name
        self.columns_map = columns_map
        self.accepted_std_div = accepted_std_div
        self.threshold = threshold


class ChecksHandler:
    def __init__(self, checks: List[OutlierCheck], conn_type: str):
        self.checks = checks
        self.conn_type = conn_type

    def prepare_main_stats_postgres_sql(self, main_table: Table):
        select_expressions = []
        for check in self.checks:
            for key in check.columns_map:
                select_expressions.append(
                    sql.SQL(
                        "stddev({key}) as {check_name}_{key_name}_stddev, avg({key}) as {check_name}_{key_name}_avg"
                    ).format(
                        key=sql.Identifier(key),
                        key_name=sql.SQL(key),
                        check_name=sql.SQL(check.name),
                    )
                )

        return sql.SQL(
            "(SELECT {select_expressions} FROM {main_table}) as main_stats"
        ).format(
            select_expressions=sql.SQL(" ,").join(select_expressions),
            main_table=sql.Identifier(main_table.table_name),
        )

    def prepare_main_stats_snowflake_sql(self, main_table: Table):
        select_expressions = []
        for check in self.checks:
            for key in check.columns_map:
                stats_query = "stddev({key}) as {check_name}_{key}_stddev, avg({key}) as {check_name}_{key}_avg"
                replacements = {"{key}": key, "{check_name}": check.name}

                invalid_identifier = is_valid_snow_identifiers(
                    [key, check.name, check.name]
                )
                if len(invalid_identifier) > 0:
                    raise ValueError(
                        f"Not a valid snowflake identifier {', '.join(invalid_identifier)}"
                    )
                for key, val in replacements.items():
                    stats_query = stats_query.replace(key, val)
                select_expressions.append(stats_query)

        if not is_valid_snow_identifier(main_table.table_name):
            raise ValueError(
                f"Not a valid snowflake identifier {main_table.table_name}"
            )
        statement = "(SELECT {select_expressions} FROM Identifier('{main_table}')) as main_stats"
        replacements = {
            "{main_table}": main_table.table_name,
            "{select_expressions}": " ,".join(select_expressions),
        }
        for key, val in replacements.items():
            statement = statement.replace(key, val)
        return statement

    def prepare_column_postgres_sql(self, check):
        column_sql = []
        for column in check.columns_map:
            main_table_col = column
            compare_table_col = check.columns_map[column]
            statement = sql.SQL(
                """({compare_table_col} > (main_stats.{check_name}_{main_table_col}_avg + (main_stats.{check_name}_{main_table_col}_stddev * {accepted_std_div})))
                OR
                ({compare_table_col} < (main_stats.{check_name}_{main_table_col}_avg - (main_stats.{check_name}_{main_table_col}_stddev * {accepted_std_div})))
                """
            ).format(
                compare_table_col=sql.Identifier(compare_table_col),
                check_name=sql.SQL(check.name),
                main_table_col=sql.SQL(main_table_col),
                accepted_std_div=sql.SQL(str(check.accepted_std_div)),
            )
            column_sql.append(statement)

        return sql.SQL("OR").join(column_sql)

    def prepare_column_snowflake_sql(self, check):
        column_sql = []
        for column in check.columns_map:
            main_table_col = column
            compare_table_col = check.columns_map[column]
            replacements = {
                "{compare_table_col}": compare_table_col,
                "{check_name}": check.name,
                "{main_table_col}": main_table_col,
                "{accepted_std_div}": str(check.accepted_std_div),
            }

            invalid_identifier = is_valid_snow_identifiers(
                [compare_table_col, check.name, main_table_col]
            )
            if len(invalid_identifier) > 0:
                raise ValueError(
                    f"Not a valid snowflake identifier {', '.join(invalid_identifier)}"
                )

            statement = """(Identifier('{compare_table_col}') > (Identifier('main_stats.{check_name}_{main_table_col}_avg') + (Identifier('main_stats.{check_name}_{main_table_col}_stddev') * {accepted_std_div})))
                OR
                (Identifier('{compare_table_col}') < (Identifier('main_stats.{check_name}_{main_table_col}_avg') - (Identifier('main_stats.{check_name}_{main_table_col}_stddev') * {accepted_std_div})))
                """
            for key, val in replacements.items():
                statement = statement.replace(key, val)
            column_sql.append(statement)

        return "OR".join(column_sql)

    def prepare_cases_postgres_sql(self):
        cases = []
        for check in self.checks:
            statement = sql.SQL(
                "CASE WHEN {column_sql} THEN 1 ELSE 0 END as {check_name}"
            ).format(
                column_sql=self.prepare_column_postgres_sql(check),
                check_name=sql.SQL(check.name),
            )
            cases.append(statement)
        return sql.SQL(",").join(cases)

    def prepare_cases_snowflake_sql(self):
        cases = []
        for check in self.checks:
            statement = "CASE WHEN {column_sql} THEN 1 ELSE 0 END as {check_name}"
            replacements = {
                "{column_sql}": self.prepare_column_snowflake_sql(check),
                "{check_name}": check.name,
            }
            for key, val in replacements.items():
                statement = statement.replace(key, val)
            cases.append(statement)
        return ",".join(cases)

    def prepare_checks_postgres_aggregation(self):
        aggregations = []
        # sum(room_check) as room_check,
        for check in self.checks:
            aggregations.append(
                sql.SQL("sum({name}) as {name}").format(name=sql.Identifier(check.name))
            )
        return sql.SQL(", ").join(aggregations)

    def prepare_checks_snowflake_aggregation(self):
        aggregations = []
        # sum(room_check) as room_check,
        for check in self.checks:
            statement = "sum({name}) as {name}"
            replacements = {"{name}": check.name}
            for key, val in replacements.items():
                statement = statement.replace(key, val)

            aggregations.append(statement)
        return ", ".join(aggregations)

    def prepare_comparison_sql(self, main_table: Table, compare_table: Table):
        statement = ""
        if self.conn_type == "postgres":
            hook = PostgresHook(postgres_conn_id=main_table.conn_id)
            statement = (
                sql.SQL(
                    "SELECT count(*) as total, {checks} FROM (SELECT {cases}, {compare_table}.*  FROM {main_stats}, {compare_table}) as results"
                )
                .format(
                    cases=self.prepare_cases_postgres_sql(),
                    main_stats=self.prepare_main_stats_postgres_sql(main_table),
                    compare_table=sql.Identifier(compare_table.table_name),
                    checks=self.prepare_checks_postgres_aggregation(),
                )
                .as_string(hook.get_conn())
            )
        elif self.conn_type == "snowflake":
            statement = "SELECT count(*) as total, {checks} FROM (SELECT {cases}, {compare_table}.*  FROM {main_stats}, Identifier('{compare_table}')) as results"
            replacements = {
                "{cases}": self.prepare_cases_snowflake_sql(),
                "{compare_table}": compare_table.table_name,
                "{main_stats}": self.prepare_main_stats_snowflake_sql(main_table),
                "{checks}": self.prepare_checks_snowflake_aggregation(),
            }

            invalid_identifier = is_valid_snow_identifiers([compare_table.table_name])
            if len(invalid_identifier) > 0:
                raise ValueError(
                    f"Not a valid snowflake identifier {', '.join(invalid_identifier)}"
                )

            for key, val in replacements.items():
                statement = statement.replace(key, val)
        else:
            raise ValueError("Unsupported connection type")
        return statement, {}

    def check_snowflake_results(self, row, check, index):
        return row[check.name.upper()]

    def check_postgres_results(self, row, check, index):
        return row[index + 1]

    def evaluate_results(self, rows, max_rows_returned, conn_type):
        check_results = {
            "postgres": self.check_postgres_results,
            "snowflake": self.check_snowflake_results,
        }[conn_type]

        total_rows = 0
        if conn_type == "snowflake":
            total_rows = rows[0]["TOTAL"]
        elif conn_type == "postgres":
            total_rows = rows[0][0]

        failed_rows_count = {
            check.name: check_results(rows[0], check, index)
            for index, check in enumerate(self.checks)
        }
        failed_checks = set()
        for check in self.checks:
            if failed_rows_count[check.name] / total_rows > check.threshold:
                failed_checks.add(check.name)
        return failed_checks

    def prepare_failed_checks_results(
        self,
        conn_type,
        main_table: Table,
        compare_table: Table,
        failed_checks: Set[str],
        max_rows_returned,
    ):
        failed_checks_sql: List[tuple] = []
        for check in self.checks:
            if check.name in failed_checks:
                if conn_type == "postgres":
                    hook = PostgresHook(postgres_conn_id=main_table.conn_id)
                    statement = (
                        sql.SQL(
                            "SELECT {compare_table}.* FROM {main_stats}, {compare_table} WHERE {column_sql} limit {max_rows_returned}"
                        )
                        .format(
                            compare_table=sql.Identifier(compare_table.table_name),
                            main_stats=self.prepare_main_stats_postgres_sql(main_table),
                            column_sql=self.prepare_column_postgres_sql(check),
                            max_rows_returned=sql.SQL(str(max_rows_returned)),
                        )
                        .as_string(hook.get_conn())
                    )
                    failed_checks_sql.append((check.name, statement, {}))
                elif conn_type == "snowflake":
                    statement = "SELECT {compare_table}.* FROM {main_stats}, Identifier('{compare_table}') WHERE {column_sql} limit {max_rows_returned}"
                    replacements = {
                        "{compare_table}": compare_table.table_name,
                        "{main_stats}": self.prepare_main_stats_snowflake_sql(
                            main_table
                        ),
                        "{column_sql}": self.prepare_column_snowflake_sql(check),
                        "{max_rows_returned}": str(max_rows_returned),
                    }

                    invalid_identifier = is_valid_snow_identifiers(
                        [compare_table.table_name]
                    )
                    if len(invalid_identifier) > 0:
                        raise ValueError(
                            f"Not a valid snowflake identifier {', '.join(invalid_identifier)}"
                        )

                    for key, val in replacements.items():
                        statement = statement.replace(key, val)
                    failed_checks_sql.append((check.name, statement, {}))
        return failed_checks_sql


class AgnosticStatsCheck(SqlDecoratoratedOperator):
    def __init__(
        self,
        checks: List[OutlierCheck],
        main_table: Table,
        compare_table: Table,
        max_rows_returned: int,
        **kwargs,
    ):
        """
        Operator to run Statistical checks on tables.

        :param main_table: main table
        :type main_table: table object
        :param compare_table: table to be compared
        :type compare_table: table object
        :param checks: check class object, which represent boolean expression
        :type checks: Check
        :param max_rows_returned: number of row returned if the check fails.
        :type max_rows_returned: int
        """

        self.main_table = main_table
        self.compare_table = compare_table
        self.max_rows_returned = max_rows_returned
        self.conn_id = main_table.conn_id
        self.checks = checks
        self.database = main_table.database

        task_id = main_table.table_name + "_" + "stats_check"

        def null_function():
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            conn_id=main_table.conn_id,
            database=main_table.database,
            schema=main_table.schema,
            warehouse=main_table.warehouse,
            task_id=task_id,
            op_args=(),
            python_callable=null_function,
            **kwargs,
        )

    def execute(self, context: Dict):
        conn_type = BaseHook.get_connection(self.conn_id).conn_type  # type: ignore
        self.handler = lambda curr: curr.fetchall()
        checkHandler = ChecksHandler(self.checks, conn_type)
        self.sql, self.parameters = checkHandler.prepare_comparison_sql(
            main_table=self.main_table, compare_table=self.compare_table
        )
        results = super().execute(context)
        failed_checks = checkHandler.evaluate_results(
            results, self.max_rows_returned, conn_type
        )
        if len(failed_checks) > 0:
            failed_check_sql = checkHandler.prepare_failed_checks_results(
                conn_type=conn_type,
                main_table=self.main_table,
                compare_table=self.compare_table,
                failed_checks=failed_checks,
                max_rows_returned=self.max_rows_returned,
            )
            failed_values = {}
            for check_query in failed_check_sql:
                check, self.sql, self.parameters = check_query
                results = super().execute(context)
                failed_values[check] = results

            raise ValueError("Stats Check Failed. {}".format(failed_values))


def stats_check(
    main_table: Table,
    compare_table: Table,
    checks: List[OutlierCheck] = [],
    max_rows_returned: int = 100,
):
    """
    :param main_table: main table
    :type main_table: table object
    :param compare_table: table to be compared
    :type compare_table: table object
    :param checks: check class object, which represent boolean expression
    :type checks: Check
    :param max_rows_returned: number of row returned if the check fails.
    :type max_rows_returned: int
    """

    return AgnosticStatsCheck(
        main_table=main_table,
        compare_table=compare_table,
        checks=checks,
        max_rows_returned=max_rows_returned,
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Callable, Iterable, List, Mapping, Optional, Union

from astro.sql.operators.agnostic_aggregate_check import aggregate_check
from astro.sql.operators.agnostic_boolean_check import boolean_check
from astro.sql.operators.agnostic_load_file import load_file
from astro.sql.operators.agnostic_save_file import save_file
from astro.sql.operators.agnostic_sql_append import SqlAppendOperator
from astro.sql.operators.agnostic_sql_merge import SqlMergeOperator
from astro.sql.operators.agnostic_sql_truncate import SqlTruncateOperator
from astro.sql.operators.agnostic_stats_check import OutlierCheck, stats_check
from astro.sql.operators.sql_decorator import (
    SqlDecoratoratedOperator,
    transform_decorator,
)
from astro.sql.parsers.sql_directory_parser import render
from astro.sql.table import Table
from astro.utils.task_id_helper import get_task_id


def transform(
    python_callable: Optional[Callable] = None,
    multiple_outputs: Optional[bool] = None,
    conn_id: str = "",
    autocommit: bool = False,
    parameters: Optional[Union[Mapping, Iterable]] = None,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    warehouse: Optional[str] = None,
):
    return transform_decorator(
        python_callable=python_callable,
        multiple_outputs=multiple_outputs,
        conn_id=conn_id,
        autocommit=autocommit,
        parameters=parameters,
        database=database,
        schema=schema,
        warehouse=warehouse,
    )


def transform_file(
    sql=None,
    conn_id: str = "",
    parameters=None,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    warehouse: Optional[str] = None,
    output_table: Table = None,
):
    def transform_file():
        return sql

    return SqlDecoratoratedOperator(
        task_id=get_task_id("transform_file", sql),
        sql=sql,
        python_callable=transform_file,
        op_kwargs={"output_table": output_table},
        op_args=(),
        conn_id=conn_id,
        parameters=parameters,
        database=database,
        schema=schema,
        warehouse=warehouse,
    )


def run_raw_sql(
    python_callable: Optional[Callable] = None,
    multiple_outputs: Optional[bool] = None,
    conn_id: str = "",
    autocommit: bool = False,
    parameters: Optional[Union[Mapping, Iterable]] = None,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    warehouse: Optional[str] = None,
):
    return transform_decorator(
        python_callable=python_callable,
        multiple_outputs=multiple_outputs,
        conn_id=conn_id,
        autocommit=autocommit,
        parameters=parameters,
        database=database,
        schema=schema,
        warehouse=warehouse,
        raw_sql=True,
    )


def append(
    append_table: Table,
    main_table: Table,
    columns: List[str] = [],
    casted_columns: dict = {},
    **kwargs,
):
    return SqlAppendOperator(
        main_table=main_table,
        append_table=append_table,
        columns=columns,
        casted_columns=casted_columns,
        **kwargs,
    )


def merge(
    target_table: Table,
    merge_table: Table,
    merge_keys: Union[List, dict],
    target_columns: List[str],
    merge_columns: List[str],
    conflict_strategy: str,
    **kwargs,
):
    """`
    :param target_table: The primary table that we are merging into
    :param merge_table: The table that will be inserted
    :param merge_keys: A key dictionary of what fields we want to compare when determining conflicts.
        ``{"foo": "bar"}`` would be equivalent to ``main_table.foo=merge_table.bar``
    :param target_columns: The columns that will be merged into (order matters and needs to be
    same length as merge_columns
    :param merge_columns:
    :param conn_id: connection ID for SQL instance
    :param conflict_strategy: Do we ignore new values on conflict or overwrite? Two strategies are "ignore" and "update"
    :param database:
    :param schema: Snowflake, specific. Specify Snowflake schema
    :param kwargs:
    :return:
    """

    return SqlMergeOperator(
        target_table=target_table,
        merge_table=merge_table,
        merge_keys=merge_keys,
        target_columns=target_columns,
        merge_columns=merge_columns,
        conflict_strategy=conflict_strategy,
    )


def truncate(
    table: Table,
    conn_id: str = "",
    database: str = "",
    schema: str = "",
    warehouse: str = "",
    **kwargs,
):
    """`
    :param table: The table that we will truncate
    :param database:
    :param schema: Snowflake, specific. Specify Snowflake schema
    :param kwargs:
    :return:
    """

    return SqlTruncateOperator(
        table=table,
        conn_id=conn_id,
        database=database,
        schema=schema,
        warehouse=warehouse,
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from numbers import Number
from typing import Dict, Optional

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from sqlalchemy.sql.functions import count
from sqlalchemy.sql.schema import Table

from astro import sql
from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.utils.task_id_helper import get_unique_task_id


class AgnosticAggregateCheck(SqlDecoratoratedOperator):
    def __init__(
        self,
        table: Table,
        check: str,
        greater_than: float = None,
        less_than: float = None,
        equal_to: float = None,
        **kwargs
    ):
        """Validate that a table has expected aggregation value.
        Range specified by greater_than and/or less_than is inclusive - [greater_than, less_than] or
        equal_to can be used to check a value.

        :param table: table name
        :type table: str
        :param check: SQL statement
        :type check: str
        :param greater_than: min expected value
        :type greater_than: Number
        :param less_than: max expected value
        :type less_than: Number
        :param equal_to: expected value
        :type equal_to: Number
        :param conn_id: connection id
        :type conn_id: str
        :param database: database name
        :type database: str
        """
        self.table = table
        self.check = check
        self.greater_than = greater_than
        self.less_than = less_than
        self.equal_to = equal_to

        if less_than is None and greater_than is None and equal_to is None:
            raise ValueError(
                "Please provide one or more of these options: less_than, greater_than, equal_to"
            )

        if (
            less_than is not None
            and greater_than is not None
            and less_than < greater_than
        ):
            raise ValueError(
                "less_than should be greater than or equal to greater_than."
            )

        task_id = get_unique_task_id("aggregate_check")

        def null_function():
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            conn_id=table.conn_id,
            task_id=task_id,
            op_args=(),
            python_callable=null_function,
            database=table.database,
            **kwargs,
        )

    def execute(self, context: Dict):
        self.sql = self.check
        self.handler = lambda curr: curr.fetchone()
        query_result = super().execute(context)
        if len(query_result) != 1:
            raise ValueError(
                "The aggregate check query should only return a single numeric value."
            )

        query_result = query_result[0]
        if not isinstance(query_result, int) and not isinstance(query_result, float):
            raise ValueError(
                "The aggregate check query should only return a single numeric value."
            )

        if self.equal_to is not None and self.equal_to != query_result:
            raise ValueError(
                "Check Failed: query result value {} not equal to {}.".format(
                    query_result, self.equal_to
                )
            )
        elif (
            self.less_than is not None
            and self.greater_than is not None
            and (self.greater_than > query_result or query_result > self.less_than)
        ):
            raise ValueError(
                "Check Failed: query result value {} not in range from {} to {}.".format(
                    query_result, self.greater_than, self.less_than
                )
            )
        elif self.less_than is not None and self.less_than < query_result:
            raise ValueError(
                "Check Failed: query result value {} not less than {}.".format(
                    query_result, self.less_than
                )
            )
        elif self.greater_than is not None and self.greater_than > query_result:
            raise ValueError(
                "Check Failed: query result value {} not greater than {}.".format(
                    query_result, self.greater_than
                )
            )


def aggregate_check(
    table: Table,
    check: str,
    greater_than: float = None,
    less_than: float = None,
    equal_to: float = None,
):
    """
    :param table: table name
    :type table: str
    :param check: SQL statement
    :type check: str
    :param greater_than: min expected rows
    :type greater_than: int
    :param less_than: max expected rows
    :type less_than: int
    :param conn_id: connection id,
    :type conn_id: str
    :param database: database name,
    :type database: str
    """
    return AgnosticAggregateCheck(
        table=table,
        check=check,
        greater_than=greater_than,
        less_than=less_than,
        equal_to=equal_to,
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
from typing import List

from airflow.exceptions import AirflowException

from astro.sql.table import Table


def _wrap_identifiers(sql, identifier_params):
    all_vals = re.findall("%\(.*?\)s", sql)
    mod_vals = {
        f: f"IDENTIFIER({f})" if f[2:-2] in identifier_params else f for f in all_vals
    }
    for k, v in mod_vals.items():
        sql = sql.replace(k, v)
    return sql


def wrap_identifier(inp):
    return "Identifier(%(" + inp + ")s)"


def snowflake_merge_func(
    target_table: Table,
    merge_table: Table,
    target_columns,
    merge_keys,
    merge_columns,
    conflict_strategy,
):
    statement = (
        "merge into Identifier(%(main_table)s) using Identifier(%(merge_table)s) on "
        "{merge_clauses}"
    )

    merge_target_dict = {
        "merge_clause_target_" + str(i): target_table.table_name + "." + x
        for i, x in enumerate(merge_keys.keys())
    }
    merge_append_dict = {
        "merge_clause_append_" + str(i): merge_table.table_name + "." + x
        for i, x in enumerate(merge_keys.values())
    }

    statement = fill_in_merge_clauses(merge_append_dict, merge_target_dict, statement)

    values_to_check = [target_table.table_name, merge_table.table_name]
    values_to_check.extend(target_columns)
    values_to_check.extend(merge_columns)
    for v in values_to_check:
        if not is_valid_snow_identifier(v):
            raise AirflowException(
                f"The identifier {v} is invalid. Please check to prevent SQL injection"
            )

    if conflict_strategy == "update":
        statement += " when matched then UPDATE SET {merge_vals}"
        statement = fill_in_update_statement(
            statement,
            target_table.table_name,
            merge_table.table_name,
            target_columns,
            merge_columns,
        )
    statement += (
        " when not matched then insert({target_columns}) values ({append_columns})"
    )
    statement = fill_in_append_statements(
        target_table.table_name,
        merge_table.table_name,
        statement,
        target_columns,
        merge_columns,
    )

    params = {}
    params.update(merge_target_dict)
    params.update(merge_append_dict)
    params["main_table"] = target_table
    params["merge_table"] = merge_table

    return statement, params


def fill_in_append_statements(
    main_table,
    merge_table,
    statement,
    targest_columns,
    merge_columns,
):
    statement = statement.replace(
        "{target_columns}",
        ",".join(f"{main_table}.{t}" for t in targest_columns),
    )
    statement = statement.replace(
        "{append_columns}",
        ",".join(f"{merge_table}.{t}" for t in merge_columns),
    )
    return statement


def fill_in_update_statement(
    statement, main_table, merge_table, target_columns, merge_columns
):
    merge_statement = ",".join(
        [
            f"{main_table}.{t}={merge_table}.{m}"
            for t, m in zip(target_columns, merge_columns)
        ]
    )
    statement = statement.replace("{merge_vals}", merge_statement)
    return statement


def fill_in_merge_clauses(merge_append_dict, merge_target_dict, statement):
    return statement.replace(
        "{merge_clauses}",
        " AND ".join(
            wrap_identifier(k) + "=" + wrap_identifier(v)
            for k, v in zip(merge_target_dict.keys(), merge_append_dict.keys())
        ),
    )


def is_valid_snow_identifier(name):
    """
    Because Snowflake does not allow using `Identifier` for inserts or updates, we need to make reasonable attempts to
    ensure that no one can perform a SQL injection using this method. The following method ensures that a string
    follows the expected identifier syntax https://docs.snowflake.com/en/sql-reference/identifiers-syntax.html
    @param name:
    @return:
    """
    if not 1 <= len(name) <= 255:
        return False

    name_is_quoted = name[0] == '"'
    if name_is_quoted:
        if len(name) < 2 or name[-1] != '"':
            return False  # invalid because no closing quote

        return ensure_internal_quotes_closed(name)

    else:  # not quoted
        return ensure_only_valid_characters(name)


def is_valid_snow_identifiers(names: List[str]):
    return [names for name in names if not is_valid_snow_identifier(name)]


# test code to check for validate snowflake identifier
def ensure_internal_quotes_closed(name):
    last_quoted = False
    for c in name[1:-1]:
        if last_quoted:
            if c != '"':
                return False
            last_quoted = False
        elif c == '"':
            last_quoted = True
        # any character is fair game inside a properly quoted name

    if last_quoted:
        return False  # last quote was not escape

    return True


def ensure_only_valid_characters(name):
    if not (name[0].isalpha()) and name[0] != "_":
        return False
    for c in name[1:]:
        if not (c.isalpha() or c.isdigit() or c == "_" or c == "$"):
            return False
    return True



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import inspect

from airflow.providers.postgres.hooks.postgres import PostgresHook
from psycopg2.extensions import AsIs

from astro.sql.table import Table


def parse_template(sql):
    return sql.replace("{", "%(").replace("}", ")s")


def process_params(parameters, python_callable):
    param_types = inspect.signature(python_callable).parameters
    return {
        k: (
            AsIs(v.schema + "." + v.table_name if v.schema else v.table_name)
            if param_types.get(k)
            and param_types.get(k).annotation == Table
            or type(v) == Table
            else v
        )
        for k, v in parameters.items()
    }


def create_sql_engine(postgres_conn_id, database):
    hook = PostgresHook(postgres_conn_id=postgres_conn_id, schema=database)
    engine = hook.get_sqlalchemy_engine()
    engine.url.database = database
    return engine



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""A decorator that allows users to run SQL queries natively in Airflow."""

__version__ = "0.3.1"
from astro.dataframe import dataframe


# This is needed to allow Airflow to pick up specific metadata fields it needs
# for certain features. We recognize it's a bit unclean to define these in
# multiple places, but at this point it's the only workaround if you'd like
# your custom conn type to show up in the Airflow UI.
def get_provider_info():
    return {
        # Required.
        "package-name": "astro-projects",
        "name": "Astro SQL Provider",
        "description": __doc__,
        "versions": [__version__],
        # Optional.
        "hook-class-names": [],
        "extra-links": [],
    }



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, List

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table
from astro.utils.postgres_append import postgres_append_func
from astro.utils.snowflake_append import snowflake_append_func
from astro.utils.task_id_helper import get_unique_task_id


class SqlAppendOperator(SqlDecoratoratedOperator):
    template_fields = ("main_table", "append_table")

    def __init__(
        self,
        append_table: Table,
        main_table: Table,
        columns: List[str] = [],
        casted_columns: dict = {},
        **kwargs,
    ):
        self.append_table = append_table
        self.main_table = main_table
        self.sql = ""

        self.columns = columns
        self.casted_columns = casted_columns
        task_id = get_unique_task_id("append_table")

        def null_function():
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            task_id=kwargs.get("task_id") or task_id,
            database=main_table.database,
            schema=main_table.schema,
            warehouse=main_table.warehouse,
            conn_id=main_table.conn_id,
            op_args=(),
            python_callable=null_function,
            **kwargs,
        )

    def execute(self, context: Dict):
        conn_type = BaseHook.get_connection(self.conn_id).conn_type  # type: ignore
        if conn_type == "postgres":
            self.sql = postgres_append_func(
                main_table=self.main_table,
                append_table=self.append_table,
                columns=self.columns,
                casted_columns=self.casted_columns,
                conn_id=self.conn_id,
            )
        elif conn_type == "snowflake":
            self.sql, self.parameters = snowflake_append_func(
                main_table=self.main_table,
                append_table=self.append_table,
                columns=self.columns,
                casted_columns=self.casted_columns,
                snowflake_conn_id=self.conn_id,
            )
        else:
            raise AirflowException(f"Please specify a postgres or snowflake conn id.")

        super().execute(context)



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Callable, Optional

from airflow.decorators.base import task_decorator_factory

from astro.sql.operators.sql_dataframe import SqlDataframeOperator


def dataframe(
    python_callable: Optional[Callable] = None,
    multiple_outputs: Optional[bool] = None,
    conn_id: str = "",
    database: Optional[str] = None,
    schema: Optional[str] = None,
    warehouse: Optional[str] = None,
):
    """
    This function allows a user to run python functions in Airflow but with the huge benefit that SQL files
    will automatically be turned into dataframes and resulting dataframes can automatically used in astro.sql functions
    """
    return task_decorator_factory(
        python_callable=python_callable,
        multiple_outputs=multiple_outputs,
        decorated_operator_class=SqlDataframeOperator,  # type: ignore
        **{
            "conn_id": conn_id,
            "database": database,
            "schema": schema,
            "warehouse": warehouse,
        }
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from typing import Union
from urllib.parse import urlparse

import pandas as pd
from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator
from smart_open import open

from astro.sql.table import Table, TempTable, create_table_name
from astro.utils.cloud_storage_creds import gcs_client, s3fs_creds
from astro.utils.load_dataframe import move_dataframe_to_sql
from astro.utils.schema_util import get_schema
from astro.utils.task_id_helper import get_task_id


class AgnosticLoadFile(BaseOperator):
    """Load S3/local table to postgres/snowflake database.

    :param path: File path.
    :type path: str
    :param output_table_name: Name of table to create.
    :type output_table_name: str
    :param file_conn_id: Airflow connection id of input file (optional)
    :type file_conn_id: str
    :param output_conn_id: Database connection id.
    :type output_conn_id: str
    """

    def __init__(
        self,
        path,
        output_table: Union[TempTable, Table],
        file_conn_id="",
        chunksize=None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.output_table: Union[TempTable, Table] = output_table
        self.path = path
        self.chunksize = chunksize
        self.file_conn_id = file_conn_id
        self.kwargs = kwargs
        self.output_table = output_table

    def execute(self, context):
        """Loads csv/parquet table from local/S3/GCS with Pandas.

        Infers SQL database type based on connection then loads table to db.
        """

        # Read file with Pandas load method based on `file_type` (S3 or local).
        df = self._load_dataframe(self.path)

        # Retrieve conn type
        conn = BaseHook.get_connection(self.output_table.conn_id)
        if type(self.output_table) == TempTable:
            self.output_table = self.output_table.to_table(
                create_table_name(context=context), get_schema()
            )
        else:
            self.output_table.schema = self.output_table.schema or get_schema()
        move_dataframe_to_sql(
            output_table_name=self.output_table.table_name,
            conn_id=self.output_table.conn_id,
            database=self.output_table.database,
            warehouse=self.output_table.warehouse,
            schema=self.output_table.schema,
            df=df,
            conn_type=conn.conn_type,
            user=conn.login,
        )
        self.log.info(f"returning table {self.output_table}")
        return self.output_table

    @staticmethod
    def validate_path(path):
        """Validate a URL or local file path"""
        try:
            result = urlparse(path)
            return all([result.scheme, result.netloc]) or os.path.isfile(path)
        except:
            return False

    def _load_dataframe(self, path):
        """Read file with Pandas.

        Select method based on `file_type` (S3 or local).
        """

        if not AgnosticLoadFile.validate_path(path):
            raise ValueError("Invalid path: {}".format(path))

        file_type = path.split(".")[-1]
        transport_params = {
            "s3": s3fs_creds,
            "gs": gcs_client,
            "": lambda: None,
        }[urlparse(path).scheme]()
        deserialiser = {
            "parquet": pd.read_parquet,
            "csv": pd.read_csv,
            "json": pd.read_json,
            "ndjson": pd.read_json,
        }
        mode = {"parquet": "rb"}
        deserialiser_params = {"ndjson": {"lines": True}}
        with open(
            path, mode=mode.get(file_type, "r"), transport_params=transport_params
        ) as stream:
            return deserialiser[file_type](
                stream, **deserialiser_params.get(file_type, {})
            )


def load_file(
    path,
    output_table=None,
    file_conn_id=None,
    task_id=None,
    **kwargs,
):
    """Convert AgnosticLoadFile into a function.

    Returns an XComArg object.

    :param path: File path.
    :type path: str
    :param output_table: Table to create
    :type output_table: Table
    :param file_conn_id: Airflow connection id of input file (optional)
    :type file_conn_id: str
    :param task_id: task id, optional.
    :type task_id: str
    """

    task_id = task_id if task_id is not None else get_task_id("load_file", path)

    return AgnosticLoadFile(
        task_id=task_id,
        path=path,
        output_table=output_table,
        file_conn_id=file_conn_id,
        **kwargs,
    ).output



import os

from airflow.providers.postgres.hooks.postgres import PostgresHook
from psycopg2 import sql


def set_schema_query(conn_type, hook, schema_id, user):

    if conn_type == "postgres":
        return (
            sql.SQL("CREATE SCHEMA IF NOT EXISTS {schema} AUTHORIZATION {user}")
            .format(schema=sql.Identifier(schema_id), user=sql.Identifier(user))
            .as_string(hook.get_conn())
        )
    elif conn_type == "snowflake":
        return f"CREATE SCHEMA IF NOT EXISTS {schema_id}"


def get_schema():
    return os.getenv("AIRFLOW__ASTRO__SQL_SCHEMA") or "tmp_astro"



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Union

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table, TempTable
from astro.utils.postgres_merge_func import postgres_merge_func
from astro.utils.snowflake_merge_func import snowflake_merge_func
from astro.utils.task_id_helper import get_unique_task_id


class SqlMergeOperator(SqlDecoratoratedOperator):
    template_fields = ("target_table", "merge_table")

    def __init__(
        self,
        target_table: Table,
        merge_table: Table,
        merge_keys,
        target_columns,
        merge_columns,
        conflict_strategy,
        **kwargs,
    ):
        self.target_table = target_table
        self.merge_table = merge_table
        self.merge_keys = merge_keys
        self.target_columns = target_columns
        self.merge_columns = merge_columns
        self.conflict_strategy = conflict_strategy
        task_id = get_unique_task_id("merge_table")

        def null_function():
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            task_id=kwargs.get("task_id") or task_id,
            op_args=(),
            python_callable=null_function,
            database=self.target_table.database,
            conn_id=self.target_table.conn_id,
            schema=self.target_table.schema,
            warehouse=self.target_table.warehouse,
            **kwargs,
        )

    def execute(self, context: Dict):
        conn_type = BaseHook.get_connection(self.conn_id).conn_type  # type: ignore
        if conn_type == "postgres":
            self.sql = postgres_merge_func(
                target_table=self.target_table,
                merge_table=self.merge_table,
                merge_keys=self.merge_keys,
                target_columns=self.target_columns,
                merge_columns=self.merge_columns,
                conflict_strategy=self.conflict_strategy,
                conn_id=self.conn_id,
            )
        elif conn_type == "snowflake":
            self.sql, self.parameters = snowflake_merge_func(
                target_table=self.target_table,
                merge_table=self.merge_table,
                merge_keys=self.merge_keys,
                target_columns=self.target_columns,
                merge_columns=self.merge_columns,
                conflict_strategy=self.conflict_strategy,
            )
        else:
            raise AirflowException(f"please give a postgres conn id")

        super().execute(context)



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Dict

from airflow.decorators.base import get_unique_task_id

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table


class SqlTruncateOperator(SqlDecoratoratedOperator):
    def __init__(
        self,
        table: Table,
        **kwargs,
    ):
        self.sql = ""
        self.table = table

        task_id = get_unique_task_id(table.table_name + "_truncate")

        def null_function(table: Table):
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            task_id=task_id,
            op_args=(),
            op_kwargs={"table": table},
            python_callable=null_function,
            **kwargs,
        )

    def execute(self, context: Dict):
        def table_func(table: Table):
            return "TRUNCATE TABLE {table}"

        self.python_callable = table_func
        self.parameters = {"table_name": self.table.table_name}

        super().execute(context)



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Optional, Union

from pandas import DataFrame
from pandas.io.sql import SQLDatabase
from snowflake.connector.pandas_tools import write_pandas

from astro.sql.operators.temp_hooks import TempPostgresHook, TempSnowflakeHook
from astro.utils.schema_util import set_schema_query


def move_dataframe_to_sql(
    output_table_name,
    conn_id,
    database,
    schema,
    warehouse,
    conn_type,
    df: DataFrame,
    user,
    chunksize=None,
):
    # Select database Hook based on `conn` type
    hook: Union[TempPostgresHook, TempSnowflakeHook] = {  # type: ignore
        "postgres": TempPostgresHook(postgres_conn_id=conn_id, schema=database),
        "snowflake": TempSnowflakeHook(
            snowflake_conn_id=conn_id,
            database=database,
            schema=schema,
            warehouse=warehouse,
        ),
    }.get(conn_type, None)
    if not hook:
        raise ValueError("conn id needs to either snowflake or postgres")
    if database:
        hook.database = database

    schema_query = set_schema_query(
        conn_type=conn_type, hook=hook, schema_id=schema, user=user
    )
    hook.run(schema_query)
    if conn_type == "snowflake":

        db = SQLDatabase(engine=hook.get_sqlalchemy_engine())
        # make columns uppercase to prevent weird errors in snowflake
        df.columns = df.columns.str.upper()
        db.prep_table(
            df,
            output_table_name.lower(),
            schema=schema,
            if_exists="replace",
            index=False,
        )
        write_pandas(
            hook.get_conn(),
            df,
            output_table_name,
            chunk_size=chunksize,
            quote_identifiers=False,
        )
    else:
        df.to_sql(
            output_table_name,
            con=hook.get_sqlalchemy_engine(),
            schema=schema,
            if_exists="replace",
            chunksize=chunksize,
            method="multi",
            index=False,
        )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import inspect
from builtins import NotImplementedError
from typing import Callable, Dict, Iterable, Mapping, Optional, Union

import pandas as pd
from airflow.decorators.base import DecoratedOperator, task_decorator_factory
from airflow.hooks.base import BaseHook
from airflow.models import DagRun, TaskInstance
from airflow.providers.postgres.hooks.postgres import PostgresHook
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook
from airflow.utils.db import provide_session
from sqlalchemy.sql.functions import Function

from astro.sql.table import Table, create_table_name
from astro.utils import postgres_transform, snowflake_transform
from astro.utils.load_dataframe import move_dataframe_to_sql
from astro.utils.schema_util import get_schema, set_schema_query


class SqlDecoratoratedOperator(DecoratedOperator):
    def __init__(
        self,
        conn_id: Optional[str] = None,
        autocommit: bool = False,
        parameters: dict = None,
        handler: Function = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        warehouse: Optional[str] = None,
        raw_sql=False,
        sql="",
        **kwargs,
    ):
        """
        :param to_dataframe: This function allows users to pull the current staging table into a pandas dataframe.

            To use this function, please make sure that your decorated function has a parameter called ``input_df``. This
        parameter will be a pandas.Dataframe that you can modify as needed. Please note that until we implement
        spark and dask dataframes, that you should be mindful as to how large your worker is when pulling large tables.
        :param from_s3: Whether to pull from s3 into current database.

            When set to true, please include a parameter named ``s3_path`` in your TaskFlow function. When calling this
        task, you can specify any s3:// path and Airflow will
        automatically pull that file into your database using Panda's automatic data typing functionality.
        :param from_csv: Whether to pull from a local csv file into current database.

            When set to true, please include a parameter named ``csv_path`` in your TaskFlow function.
        When calling this task, you can specify any local path and Airflow will automatically pull that file into your
        database using Panda's automatic data typing functionality.
        :param kwargs:
        """
        self.raw_sql = raw_sql
        self.conn_id = conn_id
        self.autocommit = autocommit
        self.parameters = parameters
        self.database = database
        self.schema = schema
        self.handler = handler
        self.warehouse = warehouse
        self.kwargs = kwargs or {}
        self.sql = sql
        self.op_kwargs: Dict = self.kwargs.get("op_kwargs") or {}
        if self.op_kwargs.get("output_table"):
            self.output_table: Optional[Table] = self.op_kwargs.pop("output_table")
        else:
            self.output_table = None

        super().__init__(
            **kwargs,
        )

    def execute(self, context: Dict):
        self.output_schema = self.schema or get_schema()
        self._set_variables_from_first_table()

        conn = BaseHook.get_connection(self.conn_id)
        self.conn_type = conn.conn_type  # type: ignore
        self.schema = self.schema or get_schema()
        self.user = conn.login
        self.run_id = context.get("run_id")
        self.convert_op_arg_dataframes()
        self.convert_op_kwarg_dataframes()
        if not self.sql:
            sql_stuff = self.python_callable(*self.op_args, **self.op_kwargs)
            # If we return two things, assume the second thing is the params
            if len(sql_stuff) == 2:
                self.sql, self.parameters = sql_stuff
            else:
                self.sql = sql_stuff
                self.parameters = {}
        elif self.sql[-4:] == ".sql":
            with open(self.sql) as file:
                self.sql = file.read().replace("\n", " ")
        if context:
            self.sql = self.render_template(self.sql, context)
            self.parameters = {
                k: self.render_template(v, context) for k, v in self.parameters.items()  # type: ignore
            }
        self._parse_template()
        output_table_name = None

        if not self.raw_sql:
            # Create a table name for the temp table
            self._set_schema_if_needed()

            if not self.output_table:
                output_table_name = create_table_name(context=context)
            else:
                output_table_name = self.output_table.table_name
            output_table_name = self.handle_schema(output_table_name)
            self.sql = self.create_temporary_table(self.sql, output_table_name)

        # Automatically add any kwargs going into the function
        if self.op_kwargs:
            self.parameters.update(self.op_kwargs)  # type: ignore

        if self.op_args:
            params = list(inspect.signature(self.python_callable).parameters.keys())
            for i, arg in enumerate(self.op_args):
                self.parameters[params[i]] = arg  # type: ignore

        self.parameters.update(self.op_kwargs)  # type: ignore

        self._process_params()
        query_result = self._run_sql(self.sql, self.parameters)
        # Run execute function of subclassed Operator.

        if self.output_table:
            self.log.info(f"returning table {self.output_table}")
            return self.output_table

        elif self.raw_sql:
            return query_result
        else:
            self.output_table = Table(
                table_name=output_table_name,
                conn_id=self.conn_id,
                database=self.database,
                warehouse=self.warehouse,
                schema=get_schema(),
            )
            self.log.info(f"returning table {self.output_table}")
            return self.output_table

    def handle_schema(self, output_table_name):
        """
        In postgres, we set the schema in the query itself instead of as a query parameter.
        This function adds the necessary {schema}.{table} notation.
        :param output_table_name:
        :return:
        """
        if self.conn_type == "postgres" and self.schema:
            output_table_name = self.schema + "." + output_table_name
        return output_table_name

    def _set_variables_from_first_table(self):
        """
        When we create our SQL operation, we run with the assumption that the first table given is the "main table".
        This means that a user doesn't need to define default conn_id, database, etc. in the function unless they want
        to create default values.
        """
        first_table: Optional[Table] = None
        if self.op_args:
            table_index = [x for x, t in enumerate(self.op_args) if type(t) == Table]
            if table_index:
                first_table = self.op_args[table_index[0]]
        if not first_table:
            table_kwargs = [
                x
                for x in inspect.signature(self.python_callable).parameters.values()
                if x.annotation == Table and type(self.op_kwargs[x.name]) == Table
            ]
            if table_kwargs:
                first_table = self.op_kwargs[table_kwargs[0].name]
        if first_table:
            self.conn_id = first_table.conn_id or self.conn_id
            self.database = first_table.database or self.database
            self.schema = first_table.schema or self.schema
            self.warehouse = first_table.warehouse or self.warehouse

    def _set_schema_if_needed(self):
        schema_statement = ""
        if self.conn_type == "postgres":
            self.hook = PostgresHook(
                postgres_conn_id=self.conn_id, schema=self.database
            )
            schema_statement = set_schema_query(
                conn_type=self.conn_type,
                hook=self.hook,
                schema_id=self.schema,
                user=self.user,
            )
        elif self.conn_type == "snowflake":
            hook = self.get_snow_hook()
            schema_statement = set_schema_query(
                conn_type=self.conn_type,
                hook=hook,
                schema_id=self.schema,
                user=self.user,
            )
        self._run_sql(schema_statement, {})

    def get_snow_hook(self) -> SnowflakeHook:
        """
        Create and return SnowflakeHook.
        :return: a SnowflakeHook instance.
        :rtype: SnowflakeHook
        """
        return SnowflakeHook(
            snowflake_conn_id=self.conn_id,
            warehouse=self.warehouse,
            database=self.database,
            role=None,
            schema=self.schema,
            authenticator=None,
            session_parameters=None,
        )

    def _run_sql(self, sql, parameters):

        if self.conn_type == "postgres":
            self.log.info("Executing: %s", sql)
            self.hook = PostgresHook(
                postgres_conn_id=self.conn_id, schema=self.database
            )
            results = self.hook.run(
                sql,
                self.autocommit,
                parameters=parameters,
                handler=self.handler,
            )
            for output in self.hook.conn.notices:
                self.log.info(output)

        elif self.conn_type == "snowflake":
            self.log.info("Executing: %s", sql)
            hook = self.get_snow_hook()
            results = hook.run(sql, autocommit=self.autocommit, parameters=parameters)
            self.query_ids = hook.query_ids

        return results

    @staticmethod
    def create_temporary_table(query, output_table_name, schema=None):
        """
        Create a temp table for the current task instance. This table will be overwritten if the DAG is run again as this
        table is only ever meant to be temporary.
        :param query:
        :param table_name:
        :return:
        """
        if schema:
            output_table_name = f"{schema}.{output_table_name}"
        return f"DROP TABLE IF EXISTS {output_table_name}; CREATE TABLE {output_table_name} AS ({query});"

    @staticmethod
    def create_cte(query, table_name):
        return f"WITH {table_name} AS ({query}) SELECT * FROM {table_name};"

    @staticmethod
    def create_output_csv_path(context):
        ti: TaskInstance = context["ti"]
        dag_run: DagRun = ti.get_dagrun()
        return f"{dag_run.dag_id}_{ti.task_id}_{int(ti.execution_date.timestamp())}.csv"

    def handle_dataframe_func(self, input_table):
        raise NotImplementedError("Need to add dataframe func to class")

    @provide_session
    def pre_execute(self, context, session=None):
        """This hook is triggered right before self.execute() is called."""
        pass

    def post_execute(self, context, result=None):
        """
        This hook is triggered right after self.execute() is called.
        """
        pass

    def _table_exists_in_db(self, conn: str, table_name: str):
        """Override this method to enable sensing db."""
        raise NotImplementedError("Add _table_exists_in_db method to class")

    def _process_params(self):
        if self.conn_type == "postgres":
            self.parameters = postgres_transform.process_params(
                self.parameters, self.python_callable
            )
        elif self.conn_type == "snowflake":
            self.parameters = snowflake_transform.process_params(self.parameters)

    def _parse_template(self):
        if self.conn_type == "postgres":
            self.sql = postgres_transform.parse_template(self.sql)
        else:
            self.sql = snowflake_transform._parse_template(
                self.sql, self.python_callable
            )

    def _cleanup(self):
        """Remove DAG's objects from S3 and db."""
        # To-do
        pass

    def convert_op_arg_dataframes(self):
        final_args = []
        for i, arg in enumerate(self.op_args):
            if type(arg) == pd.DataFrame:
                output_table_name = (
                    self.dag_id
                    + "_"
                    + self.task_id
                    + "_"
                    + self.run_id
                    + f"_input_dataframe_{i}"
                )
                move_dataframe_to_sql(
                    output_table_name=output_table_name,
                    df=arg,
                    conn_type=self.conn_type,
                    conn_id=self.conn_id,
                    database=self.database,
                    schema=self.schema,
                    warehouse=self.warehouse,
                )
                final_args.append(
                    Table(
                        table_name=output_table_name,
                        conn_id=self.conn_id,
                        database=self.database,
                        schema=self.schema,
                        warehouse=self.warehouse,
                    )
                )
            else:
                final_args.append(arg)
            self.op_args = tuple(final_args)

    def convert_op_kwarg_dataframes(self):
        final_kwargs = {}
        for k, v in self.op_kwargs.items():
            if type(v) == pd.DataFrame:
                output_table_name = "_".join(
                    [self.dag_id, self.task_id, self.run_id, "input_dataframe", str(k)]
                )
                move_dataframe_to_sql(
                    output_table_name=output_table_name,
                    df=v,
                    conn_type=self.conn_type,
                    conn_id=self.conn_id,
                    database=self.database,
                    schema=self.schema,
                    warehouse=self.warehouse,
                )
                final_kwargs[k] = output_table_name
            else:
                final_kwargs[k] = v
        self.op_kwargs = final_kwargs


def _transform_task(
    python_callable: Optional[Callable] = None,
    multiple_outputs: Optional[bool] = None,
    **kwargs,
):
    """
    Python operator decorator. Wraps a function into an Airflow operator.
    Accepts kwargs for operator kwarg. Can be reused in a single DAG.

    :param python_callable: Function to decorate
    :type python_callable: Optional[Callable]
    :param multiple_outputs: if set, function return value will be
        unrolled to multiple XCom values. List/Tuples will unroll to xcom values
        with index as key. Dict will unroll to xcom values with keys as XCom keys.
        Defaults to False.
    :type multiple_outputs: bool
    """
    return task_decorator_factory(
        python_callable=python_callable,
        multiple_outputs=multiple_outputs,
        decorated_operator_class=SqlDecoratoratedOperator,  # type: ignore
        **kwargs,
    )


def transform_decorator(
    python_callable: Optional[Callable] = None,
    multiple_outputs: Optional[bool] = None,
    conn_id: str = "",
    autocommit: bool = False,
    parameters: Optional[Union[Mapping, Iterable]] = None,
    database: Optional[str] = None,
    schema: Optional[str] = None,
    warehouse: Optional[str] = None,
    raw_sql: bool = False,
):
    """
    :param python_callable:
    :param multiple_outputs:
    :param postgres_conn_id: The :ref:`postgres conn id <howto/connection:postgres>`
        reference to a specific postgres database.
    :type postgres_conn_id: str
    :param autocommit: if True, each command is automatically committed.
        (default value: False)
    :type autocommit: bool
    :param parameters: (optional) the parameters to render the SQL query with.
    :type parameters: dict or iterable
    :param database: name of database which overwrite defined one in connection
    :type database: str
    @return:
    """
    return _transform_task(
        python_callable=python_callable,
        multiple_outputs=multiple_outputs,
        conn_id=conn_id,
        autocommit=autocommit,
        parameters=parameters,
        database=database,
        schema=schema,
        warehouse=warehouse,
        raw_sql=raw_sql,
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import inspect
import re

from astro.sql.table import Table


def handle_table(t: Table):
    return t.schema + "." + t.table_name if t.schema else t.table_name


def process_params(parameters):
    return {
        k: (handle_table(v) if type(v) == Table else v) for k, v in parameters.items()
    }


def _parse_template(sql, python_callable):
    param_types = inspect.signature(python_callable).parameters
    sql = sql.replace("{", "%(").replace("}", ")s")
    all_vals = re.findall("%\(.*?\)s", sql)
    mod_vals = {
        f: f"IDENTIFIER({f})"
        if param_types.get(f[2:-2], None)
        and param_types.get(f[2:-2], None).annotation == Table
        else f
        for f in all_vals
    }
    for k, v in mod_vals.items():
        sql = sql.replace(k, v)
    return sql



import os
from urllib import parse

import boto3
from google.cloud.storage import Client


def parse_s3_env_var():
    raw_data = (
        os.environ["AIRFLOW__ASTRO__CONN_AWS_DEFAULT"]
        .replace("%2F", "/")
        .replace("aws://", "")
        .replace("@", "")
        .split(":")
    )
    return [parse.unquote(r) for r in raw_data]


def s3fs_creds():
    # To-do: reuse this method from sql decorator
    """Structure s3fs credentials from Airflow connection.
    s3fs enables pandas to write to s3
    """
    # To-do: clean-up how S3 creds are passed to s3fs

    k, v = parse_s3_env_var()
    session = boto3.Session(
        aws_access_key_id=k,
        aws_secret_access_key=v,
    )
    return dict(client=session.client("s3"))


def gcs_client():
    """
    get GCS credentials for storage.
    """
    client = Client()
    return dict(client=client)



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Callable, Optional

from astro.dataframe import dataframe


def train(
    python_callable: Optional[Callable] = None,
    multiple_outputs: Optional[bool] = None,
    conn_id: str = "",
    database: Optional[str] = None,
    schema: Optional[str] = None,
    warehouse: Optional[str] = None,
):
    """
    For now this will have the same functionality as dataframe, but eventually we will add features
    that will optimize this experience for training models
    """
    return dataframe(
        python_callable,
        multiple_outputs,
        conn_id,
        database,
        schema,
        warehouse,
    )


def predict(
    python_callable: Optional[Callable] = None,
    multiple_outputs: Optional[bool] = None,
    conn_id: str = "",
    database: Optional[str] = None,
    schema: Optional[str] = None,
    warehouse: Optional[str] = None,
):
    """
    For now this will have the same functionality as dataframe, but eventually we will add features
    that will optimize this experience for deploying inference
    """
    return dataframe(
        python_callable,
        multiple_outputs,
        conn_id,
        database,
        schema,
        warehouse,
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import inspect
from typing import Dict, Optional

import pandas as pd
from airflow.decorators.base import DecoratedOperator
from airflow.hooks.base import BaseHook
from airflow.providers.postgres.hooks.postgres import PostgresHook
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook

from astro.sql.table import Table, TempTable, create_table_name
from astro.utils.load_dataframe import move_dataframe_to_sql
from astro.utils.schema_util import get_schema


class SqlDataframeOperator(DecoratedOperator):
    def __init__(
        self,
        conn_id: Optional[str] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        warehouse: Optional[str] = None,
        **kwargs,
    ):
        """
        Converts a SQL table into a dataframe. Users can then give a python function that takes a dataframe as
        one of its inputs and run that python function. Once that function has completed, the result is accessible
        via the Taskflow API.

        :param conn_id: Connection to the DB that you will pull the table from
        :param database: Database for input table
        :param schema:  schema for input table
        :param warehouse: (Snowflake) Which warehouse to use for the input table
        :param kwargs:
        """
        self.conn_id = conn_id
        self.database = database
        self.schema = schema
        self.warehouse = warehouse
        self.kwargs = kwargs or {}
        self.op_kwargs: Dict = self.kwargs.get("op_kwargs") or {}
        if self.op_kwargs.get("output_table"):
            self.output_table: Optional[Table] = self.op_kwargs.pop("output_table")
        else:
            self.output_table = None
        self.op_args = self.kwargs.get("op_args")  # type: ignore

        super().__init__(
            **kwargs,
        )

    def handle_op_args(self):
        full_spec = inspect.getfullargspec(self.python_callable)
        op_args = list(self.op_args)
        ret_args = []
        for arg in op_args:
            current_arg = full_spec.args.pop(0)
            if (
                full_spec.annotations[current_arg] == pd.DataFrame
                and type(arg) == Table
            ):
                ret_args.append(self._get_dataframe(arg))
            else:
                ret_args.append(arg)
        self.op_args = tuple(ret_args)

    def handle_op_kwargs(self):
        param_types = inspect.signature(self.python_callable).parameters
        self.op_kwargs = {
            k: self._get_dataframe(v)
            if param_types.get(k).annotation == pd.DataFrame and type(v) == Table
            else v
            for k, v in self.op_kwargs.items()
        }

    def execute(self, context: Dict):
        self.handle_op_args()
        self.handle_op_kwargs()

        ret = self.python_callable(*self.op_args, **self.op_kwargs)
        if self.output_table:
            if type(self.output_table) == TempTable:
                self.output_table = self.output_table.to_table(
                    table_name=create_table_name(context=context), schema=get_schema()
                )
            self.output_table.schema = self.output_table.schema or get_schema()
            conn = BaseHook.get_connection(self.output_table.conn_id)
            move_dataframe_to_sql(
                output_table_name=self.output_table.table_name,
                conn_id=self.output_table.conn_id,
                database=self.output_table.database,
                warehouse=self.output_table.warehouse,
                schema=self.output_table.schema,
                df=ret,
                conn_type=conn.conn_type,
                user=conn.login,
            )
            return self.output_table
        else:
            return ret

    def get_snow_hook(self, table: Table) -> SnowflakeHook:
        """
        Create and return SnowflakeHook.
        :return: a SnowflakeHook instance.
        :rtype: SnowflakeHook
        """
        return SnowflakeHook(
            snowflake_conn_id=table.conn_id,
            warehouse=table.warehouse,
            database=table.database,
            role=None,
            schema=table.schema,
            authenticator=None,
            session_parameters=None,
        )

    def _get_dataframe(self, table: Table):
        conn_type = BaseHook.get_connection(table.conn_id).conn_type
        self.log.info(f"Getting dataframe for {table}")
        if conn_type == "postgres":
            from psycopg2 import sql

            self.hook = PostgresHook(
                postgres_conn_id=table.conn_id, schema=table.database
            )
            query = (
                sql.SQL("SELECT * FROM {input_table}")
                .format(input_table=sql.Identifier(table.table_name))
                .as_string(self.hook.get_conn())
            )
            return self.hook.get_pandas_df(query)

        elif conn_type == "snowflake":
            hook = self.get_snow_hook(table)
            return hook.get_pandas_df(
                "SELECT * FROM IDENTIFIER(%(input_table)s)",
                parameters={"input_table": table.table_name},
            )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from airflow.decorators.base import get_unique_task_id


def get_task_id(prefix, path):
    """Generate unique tasks id based on the path.
    :parma prefix: prefix string
    :type prefix: str
    :param path: file path.
    :type path: str
    """
    task_id = "{}_{}".format(prefix, path.rsplit("/", 1)[-1].replace(".", "_"))
    return get_unique_task_id(task_id)



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re


def _wrap_identifiers(sql, identifier_params):
    all_vals = re.findall("%\(.*?\)s", sql)
    mod_vals = {
        f: f"IDENTIFIER({f})" if f[2:-2] in identifier_params else f for f in all_vals
    }
    for k, v in mod_vals.items():
        sql = sql.replace(k, v)
    return sql


def snowflake_append_func(
    main_table, columns, casted_columns, append_table, snowflake_conn_id
):
    def wrap_identifier(inp):
        return "Identifier(%(" + inp + ")s)"

    if columns or casted_columns:
        statement = (
            "INSERT INTO Identifier(%(main_table)s) ({main_cols}{sep}{main_casted_cols})"
            "(SELECT {fields}{sep}{casted_fields} FROM Identifier(%(append_table)s))"
        )
    else:
        statement = "INSERT INTO Identifier(%(main_table)s) (SELECT * FROM Identifier(%(append_table)s))"

    col_dict = {
        f"col{i}": x for i, x in enumerate(columns)
    }  # prevent directly putting user input into query

    casted_col_dict = {f"casted_col{i}": x for i, x in enumerate(casted_columns.keys())}

    statement = (
        statement.replace("{sep}", ",")
        if columns and casted_columns
        else statement.replace("{sep}", "")
    )
    statement = statement.replace(
        "{main_cols}", ",".join([c for c in columns])
    )  # Please note that we are not wrapping these in Identifier due to a snowflake bug. Must fix before public release!
    statement = statement.replace(
        "{main_casted_cols}", ",".join([c for c in casted_columns.keys()])
    )  # Please note that we are not wrapping these in Identifier due to a snowflake bug. Must fix before public release!
    statement = statement.replace(
        "{fields}", ",".join([wrap_identifier(c) for c in col_dict.keys()])
    )
    statement = statement.replace(
        "{casted_fields}",
        ",".join(
            [
                "CAST(" + wrap_identifier(c) + " AS " + casted_columns[v] + ")"
                for c, v in casted_col_dict.items()
            ]
        ),
    )
    statement = statement.replace("{", "%(").replace("}", ")s")
    statement = _wrap_identifiers(
        sql=statement,
        identifier_params=[main_table, columns, casted_columns.keys(), append_table],
    )
    col_dict.update(casted_col_dict)
    col_dict["main_table"] = main_table
    col_dict["append_table"] = append_table
    return statement, col_dict



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from airflow.models import DagRun, TaskInstance


class Table:
    def __init__(
        self, table_name="", conn_id=None, database=None, schema=None, warehouse=None
    ):
        self.table_name = table_name
        self.conn_id = conn_id
        self.database = database
        self.schema = schema
        self.warehouse = warehouse

    def identifier_args(self):
        return (self.schema, self.table_name) if self.schema else (self.table_name,)

    def qualified_name(self):
        return self.schema + "." + self.table_name if self.schema else self.table_name

    def __str__(self):
        return f"Table(table_name={self.table_name}, database={self.database}, schema={self.schema}, conn_id={self.conn_id}, warehouse={self.warehouse})"


class TempTable(Table):
    def __init__(self, conn_id=None, database=None, warehouse=""):
        super().__init__(
            table_name="", conn_id=conn_id, database=database, warehouse=warehouse
        )

    def to_table(self, table_name: str, schema: str) -> Table:
        return Table(
            table_name=table_name,
            conn_id=self.conn_id,
            database=self.database,
            warehouse=self.warehouse,
            schema=schema,
        )


def create_table_name(context):
    ti: TaskInstance = context["ti"]
    dag_run: DagRun = ti.get_dagrun()
    table_name = f"{dag_run.dag_id}_{ti.task_id}_{dag_run.id}".replace("-", "_")
    if not table_name.isidentifier():
        table_name = f'"{table_name}"'
    return table_name
